"""
Process manager
----------------

Process managers are used to manage run and build processes. They are responsible for
starting, stopping, updating, and checking status of processes. Processes themselves
are stored in the `processes` dictionary of process managers.
"""
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from omegaconf import OmegaConf

from chatsky_ui.core.config import settings
from chatsky_ui.core.logger_config import get_logger
from chatsky_ui.db.base import read_conf, read_logs
from chatsky_ui.schemas.preset import Preset
from chatsky_ui.schemas.process_status import Status
from chatsky_ui.services.process import BuildProcess, RunProcess


class ProcessManager:
    """Base for build and run process managers."""

    def __init__(self):
        self.processes: Dict[int, Union[BuildProcess, RunProcess]] = {}
        self.last_id: int
        self._logger = None

    @property
    def logger(self):
        if self._logger is None:
            raise ValueError("Logger has not been configured. Call set_logger() first.")
        return self._logger

    def set_logger(self):
        self._logger = get_logger(__name__)

    def get_last_id(self):
        """Gets the maximum id among processes of type BuildProcess or RunProcess."""
        return self.last_id

    async def stop(self, id_: int) -> None:
        """Stops the process with the given id.

        raises:
            ProcessLookupError: If the process with the given id is not found.
            RuntimeError: If the process has not started yet.
        """
        if id_ not in self.processes:
            self.logger.error("Process with id '%s' not found in recent running processes", id_)
            raise ProcessLookupError
        try:
            await self.processes[id_].stop()
        except (RuntimeError, ProcessLookupError):
            raise

    async def check_status(self, id_: int, *args, **kwargs) -> None:
        """Checks the status of the process with the given id by calling the `periodically_check_status`
        method of the process.

        This updates the process status in the database every 2 seconds.
        """
        await self.processes[id_].periodically_check_status()

    async def get_status(self, id_: int) -> Status:
        """Checks the status of the process with the given id by calling the `check_status` method of the process."""
        return await self.processes[id_].check_status()

    async def get_process_info(self, id_: int, path: Path) -> Optional[Dict[str, Any]]:
        """Returns metadata of a specific process identified by its unique ID."""
        db_conf = await read_conf(path)
        conf_dict = OmegaConf.to_container(db_conf, resolve=True)
        return next((db_process for db_process in conf_dict if db_process["id"] == id_), None)  # type: ignore

    async def get_full_info(self, offset: int, limit: int, path: Path) -> List[Dict[str, Any]]:
        """Returns metadata of ``limit`` number of processes, starting from the ``offset``th process."""

        db_conf = await read_conf(path)
        conf_dict = OmegaConf.to_container(db_conf, resolve=True)
        return conf_dict[offset : offset + limit]  # type: ignore

    async def fetch_process_logs(self, id_: int, offset: int, limit: int, path: Path) -> Optional[List[str]]:
        """Returns the logs of one process according to its id. If the process is not found, returns None."""
        process_info = await self.get_process_info(id_, path)
        if process_info is None:
            self.logger.error("Id '%s' not found", id_)
            return None

        log_file = Path(process_info["log_path"])
        try:
            logs = await read_logs(log_file)
            logs = [log for log in logs if log.strip()]
        except FileNotFoundError:
            self.logger.error("Log file '%s' not found", log_file)
            return None

        if offset > len(logs):
            self.logger.info("Offset '%s' is out of bounds ('%s' logs found)", offset, len(logs))
            return None  # TODO: raise error!

        self.logger.info("Returning %s logs", len(logs))
        return logs[offset : offset + limit]


class RunManager(ProcessManager):
    """Process manager for running a DFF pipeline."""

    async def start(self, build_id: int, preset: Preset) -> int:
        """Starts a new run process.

        Increases the maximum existing id by 1 and assigns it to the new process.
        Starts the process and appends it to the processes list.

        Args:
            build_id (int): the build id to run
            preset (Preset): the preset to use among ("success", "failure", "loop")

        Returns:
            int: the id of the new started process
        """
        cmd_to_run = (
            f"chatsky.ui run_bot --build-id {build_id} "
            f"--preset {preset.end_status} "
            f"--project-dir {settings.work_directory}"
        )
        self.last_id = max([run["id"] for run in await self.get_full_info(0, 10000)])
        self.last_id += 1
        id_ = self.last_id
        process = RunProcess(id_, build_id, preset.end_status)
        await process.start(cmd_to_run)
        process.logger.debug("Started process. status: '%s'", process.process.returncode)
        self.processes[id_] = process

        return self.last_id

    async def get_run_info(self, id_: int) -> Optional[Dict[str, Any]]:
        """Returns metadata of  a specific run process identified by its unique ID."""
        return await super().get_process_info(id_, settings.runs_path)

    async def get_full_info(self, offset: int, limit: int, path: Path = None) -> List[Dict[str, Any]]:
        """Returns metadata of ``limit`` number of run processes, starting from the ``offset``th process."""
        path = path or settings.runs_path
        return await super().get_full_info(offset, limit, path)

    async def fetch_run_logs(self, run_id: int, offset: int, limit: int) -> Optional[List[str]]:
        """Returns the logs of one run according to its id.

        Number of loglines returned is based on `offset` as the start line and limited by `limit` lines.
        """
        return await self.fetch_process_logs(run_id, offset, limit, settings.runs_path)


class BuildManager(ProcessManager):
    """Process manager for converting a frontned graph to a DFF script."""

    async def start(self, preset: Preset) -> int:
        """Starts a new build process.

        Increases the maximum existing id by 1 and assigns it to the new process.
        Starts the process and appends it to the processes list.

        Args:
            preset (Preset): the preset to use among ("success", "failure", "loop")

        Returns:
            int: the id of the new started process
        """
        self.last_id = max([build["id"] for build in await self.get_full_info(0, 10000)])
        self.last_id += 1
        id_ = self.last_id
        process = BuildProcess(id_, preset.end_status)
        cmd_to_run = (
            f"chatsky.ui build_bot --build-id {id_} "
            f"--preset {preset.end_status} "
            f"--project-dir {settings.work_directory}"
        )
        await process.start(cmd_to_run)
        self.processes[id_] = process

        return self.last_id

    async def check_status(self, id_, index, *args, **kwargs):
        """Checks the build "id_" process status by calling the `periodically_check_status`
        method of the process.

        This updates the process status in the database every 2 seconds.
        The index is refreshed after the build is done/failed.
        """
        await self.processes[id_].periodically_check_status()
        await index.load()

    async def get_build_info(self, id_: int, run_manager: RunManager) -> Optional[Dict[str, Any]]:
        """Returns metadata of a specific build process identified by its unique ID.

        Args:
            ``id_`` (int): the id of the build
            ``run_manager`` (RunManager): the run manager to use for getting all runs of this build
        """
        builds_info = await self.get_full_info_with_runs_info(run_manager, offset=0, limit=10**5)
        return next((build for build in builds_info if build["id"] == id_), None)

    async def get_full_info(self, offset: int, limit: int, path: Path = None) -> List[Dict[str, Any]]:
        """Returns metadata of ``limit`` number of processes, starting from the ``offset`` process."""
        path = path or settings.builds_path
        return await super().get_full_info(offset, limit, path)

    async def get_full_info_with_runs_info(
        self, run_manager: RunManager, offset: int, limit: int
    ) -> List[Dict[str, Any]]:
        """Returns metadata of ``limit`` number of processes, starting from the ``offset``th process.

        Args:
            run_manager (RunManager): the run manager to use for getting all runs of this build
        """
        builds_info = await self.get_full_info(offset=offset, limit=limit)
        runs_info = await run_manager.get_full_info(offset=0, limit=10**5)
        for build in builds_info:
            del build["run_ids"]
            build["runs"] = []
            for run in runs_info:
                if build["id"] == run["build_id"]:
                    run_without_build_id = {k: v for k, v in run.items() if k != "build_id"}
                    build["runs"].append(run_without_build_id)

        return builds_info

    async def fetch_build_logs(self, build_id: int, offset: int, limit: int) -> Optional[List[str]]:
        """Returns the logs of one build according to its id.

        Number of loglines returned is based on `offset` as the start line and limited by `limit` lines.
        """
        return await self.fetch_process_logs(build_id, offset, limit, settings.builds_path)



import asyncio
import json
import os
import string
import sys
from pathlib import Path

import nest_asyncio
import typer
from cookiecutter.main import cookiecutter
from typing_extensions import Annotated

# Patch nest_asyncio before importing DFF
nest_asyncio.apply = lambda: None

from chatsky_ui.core.config import app_runner, settings  # noqa: E402
from chatsky_ui.core.logger_config import get_logger  # noqa: E402

cli = typer.Typer(
    help="🚀 Welcome to Chatsky-UI!\n\n"
    "To get started, use the following commands:\n\n"
    "1. `init` - Initializes a new Chatsky-UI project.\n\n"
    "2. `run_app` - Runs the UI for your project.\n"
)


async def _execute_command(command_to_run):
    logger = get_logger(__name__)
    try:
        process = await asyncio.create_subprocess_exec(*command_to_run.split())

        # Check the return code to determine success
        if process.returncode == 0:
            logger.info("Command '%s' executed successfully.", command_to_run)
        elif process.returncode is None:
            logger.info("Process by command '%s' is running.", command_to_run)
            await process.wait()
            logger.info("Process ended with return code: %d.", process.returncode)
            sys.exit(process.returncode)
        else:
            logger.error("Command '%s' failed with return code: %d", command_to_run, process.returncode)
            sys.exit(process.returncode)

    except Exception as e:
        logger.error("Error executing '%s': %s", command_to_run, str(e))
        sys.exit(1)


def _execute_command_file(build_id: int, project_dir: Path, command_file: str, preset: str):
    logger = get_logger(__name__)

    presets_build_path = settings.presets / command_file
    with open(presets_build_path, encoding="UTF-8") as file:
        file_content = file.read()

    template = string.Template(file_content)
    substituted_content = template.substitute(work_directory=project_dir, build_id=build_id)

    presets_build_file = json.loads(substituted_content)
    if preset in presets_build_file:
        command_to_run = presets_build_file[preset]["cmd"]
        logger.debug("Executing command for preset '%s': %s", preset, command_to_run)

        asyncio.run(_execute_command(command_to_run))
    else:
        raise ValueError(f"Invalid preset '{preset}'. Preset must be one of {list(presets_build_file.keys())}")


@cli.command("build_bot")
def build_bot(
    build_id: Annotated[int, typer.Option(help="Id to save the build with")] = None,
    project_dir: Path = None,
    preset: Annotated[str, typer.Option(help="Could be one of: success, failure, loop")] = "success",
):
    """Builds the bot with one of three various presets."""
    project_dir = project_dir or settings.work_directory

    if not project_dir.is_dir():
        raise NotADirectoryError(f"Directory {project_dir} doesn't exist")
    settings.set_config(work_directory=project_dir)

    _execute_command_file(build_id, project_dir, "build.json", preset)


@cli.command("build_scenario")
def build_scenario(
    build_id: Annotated[int, typer.Argument(help="Id to save the build with")],
    project_dir: Annotated[Path, typer.Option(help="Your Chatsky-UI project directory")] = ".",
    # TODO: add custom_dir - maybe the same way like project_dir
):
    """Builds the bot with preset `success`"""
    if not project_dir.is_dir():
        raise NotADirectoryError(f"Directory {project_dir} doesn't exist")
    settings.set_config(work_directory=project_dir)

    from chatsky_ui.services.json_converter import converter  # pylint: disable=C0415

    asyncio.run(converter(build_id=build_id))


@cli.command("run_bot")
def run_bot(
    build_id: Annotated[int, typer.Option(help="Id of the build to run")] = None,
    project_dir: Annotated[Path, typer.Option(help="Your Chatsky-UI project directory")] = None,
    preset: Annotated[str, typer.Option(help="Could be one of: success, failure, loop")] = "success",
):
    """Runs the bot with one of three various presets."""
    project_dir = project_dir or settings.work_directory

    if not project_dir.is_dir():
        raise NotADirectoryError(f"Directory {project_dir} doesn't exist")
    settings.set_config(work_directory=project_dir)

    _execute_command_file(build_id, project_dir, "run.json", preset)


@cli.command("run_scenario")
def run_scenario(
    build_id: Annotated[int, typer.Argument(help="Id of the build to run")],
    project_dir: Annotated[Path, typer.Option(help="Your Chatsky-UI project directory")] = ".",
):
    """Runs the bot with preset `success`"""
    if not project_dir.is_dir():
        raise NotADirectoryError(f"Directory {project_dir} doesn't exist")
    settings.set_config(work_directory=project_dir)
    script_path = settings.scripts_dir / f"build_{build_id}.yaml"

    command_to_run = f"python {project_dir}/app.py --script-path {script_path}"
    asyncio.run(_execute_command(command_to_run))


@cli.command("run_app")
def run_app(
    host: str = None,
    port: int = None,
    log_level: str = None,
    conf_reload: Annotated[bool, typer.Option(help="True for dev-mode, False otherwise")] = None,
    project_dir: Annotated[Path, typer.Option(help="Your Chatsky-UI project directory")] = Path("."),
) -> None:
    """Runs the UI for your `project_dir` on `host:port`."""
    host = host or settings.host
    port = port or settings.port
    log_level = log_level or settings.log_level
    conf_reload = conf_reload or settings.conf_reload

    if not project_dir.is_dir():
        raise NotADirectoryError(f"Directory {project_dir} doesn't exist")

    settings.set_config(
        host=host,
        port=port,
        log_level=log_level,
        conf_reload=str(conf_reload).lower() in ["true", "yes", "t", "y", "1"],
        work_directory=project_dir,
    )
    if conf_reload:
        settings.save_config()  # this is for the sake of maintaining the state of the settings

    app_runner.set_settings(settings)
    app_runner.run()


@cli.command("init")
def init(
    destination: Annotated[Path, typer.Option(help="Path where you want to create your project")] = None,
    no_input: Annotated[bool, typer.Option(help="True for quick and easy initialization using default values")] = False,
    overwrite_if_exists: Annotated[
        bool,
        typer.Option(help="True for replacing any project named as `my_project`)"),
    ] = True,
):
    """Initializes a new Chatsky-UI project using an off-the-shelf template."""
    destination = destination or settings.work_directory

    original_dir = os.getcwd()
    try:
        os.chdir(destination)
        cookiecutter(
            "https://github.com/Ramimashkouk/df_d_template.git",
            no_input=no_input,
            overwrite_if_exists=overwrite_if_exists,
        )
    finally:
        os.chdir(original_dir)



import re
from io import StringIO
from typing import Dict, Optional, Union

import aiofiles
from fastapi import APIRouter, Depends
from pylint.lint import Run, pylinter
from pylint.reporters.text import TextReporter

from chatsky_ui.api.deps import get_index
from chatsky_ui.clients.dff_client import get_dff_conditions
from chatsky_ui.core.config import settings
from chatsky_ui.schemas.code_snippet import CodeSnippet
from chatsky_ui.services.index import Index
from chatsky_ui.utils.ast_utils import get_imports_from_file

router = APIRouter()


@router.get("/search/{service_name}", status_code=200)
async def search_service(service_name: str, index: Index = Depends(get_index)) -> Dict[str, Optional[Union[str, list]]]:
    """Searches for a custom service by name and returns its code.

    A service could be a condition, reponse, or pre/postservice.
    """
    response = await index.search_service(service_name)
    return {"status": "ok", "data": response}


@router.post("/lint_snippet", status_code=200)
async def lint_snippet(snippet: CodeSnippet) -> Dict[str, str]:
    """Lints a snippet with Pylint.

    This endpoint Joins the snippet with all imports existing in the conditions.py file and then runs Pylint on it.
    """
    code_snippet = snippet.code.replace(r"\n", "\n")

    imports = get_imports_from_file(settings.conditions_path)
    code_snippet = "\n\n".join([imports, code_snippet])

    async with aiofiles.open(settings.snippet2lint_path, "wt", encoding="UTF-8") as file:
        await file.write(code_snippet)

    pylint_output = StringIO()
    reporter = TextReporter(pylint_output)
    Run([str(settings.snippet2lint_path), "--disable=W,I,R,C"], reporter=reporter, exit=False)

    error = pylint_output.getvalue()
    if re.search(r": E\d{4}:", error):
        response = {"status": "error", "message": error}
    else:
        response = {"status": "ok", "message": ""}
    pylinter.MANAGER.clear_cache()
    return response


@router.get("/get_conditions", status_code=200)
async def get_conditions() -> Dict[str, Union[str, list]]:
    """Gets the dff's out-of-the-box conditions."""
    return {"status": "ok", "data": get_dff_conditions()}



import asyncio

import pytest

from chatsky_ui.schemas.process_status import Status


class TestRunProcess:
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "cmd_to_run, status",
        [
            ("sleep 10000", Status.RUNNING),
            ("false", Status.FAILED),
            ("echo Hello", Status.COMPLETED),
        ],
    )
    async def test_check_status(self, run_process, cmd_to_run, status):
        process = await run_process(cmd_to_run)
        await asyncio.sleep(2)
        assert await process.check_status() == status

    # def test_periodically_check_status(self, run_process):
    #     process = await run_process("sleep 10000")
    #     run_process.periodically_check_status()

    @pytest.mark.asyncio
    async def test_stop(self, run_process):
        process = await run_process("sleep 10000")
        await process.stop()
        assert process.process.returncode == -15

    @pytest.mark.asyncio
    async def test_read_stdout(self, run_process):
        process = await run_process("echo Hello")
        output = await process.read_stdout()
        assert output.strip().decode() == "Hello"

    @pytest.mark.asyncio
    async def test_write_stdout(self, run_process):
        process = await run_process("cat")
        await process.write_stdin(b"Chatsky-UI team welcome you.\n")
        output = await process.process.stdout.readline()
        assert output.decode().strip() == "Chatsky-UI team welcome you."


# class TestBuildProcess:
#     pass



"""
Websocket class for controling websocket operations.
"""
import asyncio
from asyncio.tasks import Task
from typing import Dict, List, Set

from fastapi import WebSocket, WebSocketDisconnect

from chatsky_ui.core.logger_config import get_logger
from chatsky_ui.services.process_manager import ProcessManager


class WebSocketManager:
    """Controls websocket operations connect, disconnect, check status, and communicate."""

    def __init__(self):
        self.pending_tasks: Dict[WebSocket, Set[Task]] = dict()
        self.active_connections: List[WebSocket] = []
        self._logger = None

    @property
    def logger(self):
        if self._logger is None:
            raise ValueError("Logger has not been configured. Call set_logger() first.")
        return self._logger

    def set_logger(self):
        self._logger = get_logger(__name__)

    async def connect(self, websocket: WebSocket):
        """Accepts the websocket connection and marks it as active connection."""
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket):
        """Cancels pending tasks of the open websocket process and removes it from active connections."""
        # TODO: await websocket.close()
        if websocket in self.pending_tasks:
            self.logger.info("Cancelling pending tasks")
            for task in self.pending_tasks[websocket]:
                task.cancel()
            del self.pending_tasks[websocket]
        self.active_connections.remove(websocket)

    def check_status(self, websocket: WebSocket):
        if websocket in self.active_connections:
            return websocket  # return Status!

    async def send_process_output_to_websocket(
        self, run_id: int, process_manager: ProcessManager, websocket: WebSocket
    ):
        """Reads and forwards process output to the websocket client."""
        try:
            while True:
                response = await process_manager.processes[run_id].read_stdout()
                if not response:
                    break
                await websocket.send_text(response.decode().strip())
        except WebSocketDisconnect:
            self.logger.info("Websocket connection is closed by client")
        except RuntimeError:
            raise

    async def forward_websocket_messages_to_process(
        self, run_id: int, process_manager: ProcessManager, websocket: WebSocket
    ):
        """Listens for messages from the websocket and sends them to the process."""
        try:
            while True:
                user_message = await websocket.receive_text()
                if not user_message:
                    break
                await process_manager.processes[run_id].write_stdin(user_message.encode() + b"\n")
        except asyncio.CancelledError:
            self.logger.info("Websocket connection is closed")
        except WebSocketDisconnect:
            self.logger.info("Websocket connection is closed by client")
        except RuntimeError:
            raise



import asyncio
from typing import Any, Dict, List, Optional, Union

from fastapi import APIRouter, BackgroundTasks, Depends, HTTPException, WebSocket, WebSocketException, status

from chatsky_ui.api import deps
from chatsky_ui.schemas.pagination import Pagination
from chatsky_ui.schemas.preset import Preset
from chatsky_ui.services.index import Index
from chatsky_ui.services.process_manager import BuildManager, ProcessManager, RunManager
from chatsky_ui.services.websocket_manager import WebSocketManager

router = APIRouter()


async def _stop_process(id_: int, process_manager: ProcessManager, process="run") -> Dict[str, str]:
    """Stops a `build` or `run` process with the given id."""

    try:
        await process_manager.stop(id_)
    except (RuntimeError, ProcessLookupError) as e:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Process not found. It may have already exited or not started yet. Please check logs.",
        ) from e

    process_manager.logger.info("%s process '%s' has stopped", process.capitalize(), id_)
    return {"status": "ok"}


async def _check_process_status(id_: int, process_manager: ProcessManager) -> Dict[str, str]:
    """Checks the status of a `build` or `run` process with the given id."""
    if id_ not in process_manager.processes:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Process not found. It may have already exited.",
        )
    process_status = await process_manager.get_status(id_)
    return {"status": process_status.value}


@router.post("/build/start", status_code=201)
async def start_build(
    preset: Preset,
    background_tasks: BackgroundTasks,
    build_manager: BuildManager = Depends(deps.get_build_manager),
    index: Index = Depends(deps.get_index),
) -> Dict[str, Union[str, int]]:

    """Starts a `build` process with the given preset.

    This runs a background task to check the status of the process every 2 seconds.

    Args:
        preset (Preset): The preset to set the build process for. Must be among ("success", "failure", "loop")

    Returns:
        {"status": "ok", "build_id": build_id}: in case of **starting** the build process successfully.
    """

    await asyncio.sleep(preset.wait_time)
    build_id = await build_manager.start(preset)
    background_tasks.add_task(build_manager.check_status, build_id, index)
    build_manager.logger.info("Build process '%s' has started", build_id)
    return {"status": "ok", "build_id": build_id}


@router.get("/build/stop/{build_id}", status_code=200)
async def stop_build(*, build_id: int, build_manager: BuildManager = Depends(deps.get_build_manager)) -> Dict[str, str]:
    """Stops a `build` process with the given id.

    Args:
        build_id (int): The id of the process to stop.
        build_id (BuildManager): The process manager dependency to stop the process with.

    Raises:
        HTTPException: With status code 404 if the process is not found.

    Returns:
        {"status": "ok"}: in case of stopping a process successfully.
    """
    return await _stop_process(build_id, build_manager, process="build")


@router.get("/build/status/{build_id}", status_code=200)
async def check_build_status(
    *, build_id: int, build_manager: BuildManager = Depends(deps.get_build_manager)
) -> Dict[str, str]:
    """Checks the status of a `build` process with the given id.

    Args:
        build_id (int): The id of the process to check.
        build_manager (BuildManager): The process manager dependency to check the process with.

    Raises:
        HTTPException: With status code 404 if the process is not found.

    Returns:
        {"status": "completed"}: in case of a successfully completed process.
        {"status": "running"}: in case of a still running process.
        {"status": "stopped"}: in case of a stopped process.
        {"status": "failed"}: in case of a failed-to-run process.
    """
    return await _check_process_status(build_id, build_manager)


@router.get("/builds", response_model=Optional[Union[list, dict]], status_code=200)
async def check_build_processes(
    build_id: Optional[int] = None,
    build_manager: BuildManager = Depends(deps.get_build_manager),
    run_manager: RunManager = Depends(deps.get_run_manager),
    pagination: Pagination = Depends(),
) -> Optional[Union[Dict[str, Any], List[Dict[str, Any]]]]:
    """Checks the status of all `build` processes and returns them along with their runs info.

    The offset and limit parameters can be used to paginate the results.

    Args:
        build_id (Optional[int]): The id of the process to check. If not specified, all processes will be returned.
    """
    if build_id is not None:
        return await build_manager.get_build_info(build_id, run_manager)
    else:
        return await build_manager.get_full_info_with_runs_info(
            run_manager, offset=pagination.offset(), limit=pagination.limit
        )


@router.get("/builds/logs/{build_id}", response_model=Optional[list], status_code=200)
async def get_build_logs(
    build_id: int, build_manager: BuildManager = Depends(deps.get_build_manager), pagination: Pagination = Depends()
) -> Optional[List[str]]:
    """Gets the logs of a specific `build` process.

    The offset and limit parameters can be used to paginate the results.
    """
    if build_id is not None:
        return await build_manager.fetch_build_logs(build_id, pagination.offset(), pagination.limit)


@router.post("/run/start/{build_id}", status_code=201)
async def start_run(
    *,
    build_id: int,
    preset: Preset,
    background_tasks: BackgroundTasks,
    run_manager: RunManager = Depends(deps.get_run_manager)
) -> Dict[str, Union[str, int]]:
    """Starts a `run` process with the given preset.

    This runs a background task to check the status of the process every 2 seconds.

    Args:
        build_id (int): The id of the build process to start running.
        preset (Preset): The preset to set the build process for. Must be among ("success", "failure", "loop")

    Returns:
        {"status": "ok", "build_id": run_id}: in case of **starting** the run process successfully.
    """

    await asyncio.sleep(preset.wait_time)
    run_id = await run_manager.start(build_id, preset)
    background_tasks.add_task(run_manager.check_status, run_id)
    run_manager.logger.info("Run process '%s' has started", run_id)
    return {"status": "ok", "run_id": run_id}


@router.get("/run/stop/{run_id}", status_code=200)
async def stop_run(*, run_id: int, run_manager: RunManager = Depends(deps.get_run_manager)) -> Dict[str, str]:
    """Stops a `run` process with the given id.

    Args:
        run_id (int): The id of the process to stop.
        run_manager (RunManager): The process manager dependency to stop the process with.

    Raises:
        HTTPException: With status code 404 if the process is not found.

    Returns:
        {"status": "ok"}: in case of stopping a process successfully.
    """

    return await _stop_process(run_id, run_manager, process="run")


@router.get("/run/status/{run_id}", status_code=200)
async def check_run_status(*, run_id: int, run_manager: RunManager = Depends(deps.get_run_manager)) -> Dict[str, Any]:
    """Checks the status of a `run` process with the given id.

    Args:
        build_id (int): The id of the process to check.
        run_manager (RunManager): The process manager dependency to check the process with.

    Raises:
        HTTPException: With status code 404 if the process is not found.

    Returns:
        {"status": "alive"}: in case of a successfully run process. Now it is able to communicate.
        {"status": "running"}: in case of a still running process.
        {"status": "stopped"}: in case of a stopped process.
        {"status": "failed"}: in case of a failed-to-run process.
    """
    return await _check_process_status(run_id, run_manager)


@router.get("/runs", response_model=Optional[Union[list, dict]], status_code=200)
async def check_run_processes(
    run_id: Optional[int] = None,
    run_manager: RunManager = Depends(deps.get_run_manager),
    pagination: Pagination = Depends(),
) -> Optional[Union[Dict[str, Any], List[Dict[str, Any]]]]:
    """Checks the status of all `run` processes and returns them.

    The offset and limit parameters can be used to paginate the results.

    Args:
        run_id (Optional[int]): The id of the process to check. If not specified, all processes will be returned.
    """

    if run_id is not None:
        return await run_manager.get_run_info(run_id)
    else:
        return await run_manager.get_full_info(offset=pagination.offset(), limit=pagination.limit)


@router.get("/runs/logs/{run_id}", response_model=Optional[list], status_code=200)
async def get_run_logs(
    run_id: int, run_manager: RunManager = Depends(deps.get_run_manager), pagination: Pagination = Depends()
) -> Optional[List[str]]:
    """Gets the logs of a specific `run` process.

    The offset and limit parameters can be used to paginate the results.
    """
    if run_id is not None:
        return await run_manager.fetch_run_logs(run_id, pagination.offset(), pagination.limit)


@router.websocket("/run/connect")
async def connect(
    websocket: WebSocket,
    websocket_manager: WebSocketManager = Depends(deps.get_websocket_manager),
    run_manager: RunManager = Depends(deps.get_run_manager),
) -> None:
    """Establishes a WebSocket connection to communicate with an alive run process identified by its 'run_id'.

    The WebSocket URL should adhere to the format: /bot/run/connect?run_id=<run_id>.
    """

    run_manager.logger.debug("Connecting to websocket")
    run_id = websocket.query_params.get("run_id")

    # Validate run_id
    if run_id is None:
        run_manager.logger.error("No run_id provided")
        raise WebSocketException(code=status.WS_1008_POLICY_VIOLATION)
    if not run_id.isdigit():
        run_manager.logger.error("A non-digit run run_id provided")
        raise WebSocketException(code=status.WS_1003_UNSUPPORTED_DATA)
    run_id = int(run_id)
    if run_id not in run_manager.processes:
        run_manager.logger.error("process with run_id '%s' exited or never existed", run_id)
        raise WebSocketException(code=status.WS_1014_BAD_GATEWAY)

    await websocket_manager.connect(websocket)
    run_manager.logger.info("Websocket for run process '%s' has been opened", run_id)

    await websocket.send_text("Start chatting")

    output_task = asyncio.create_task(
        websocket_manager.send_process_output_to_websocket(run_id, run_manager, websocket)
    )
    input_task = asyncio.create_task(
        websocket_manager.forward_websocket_messages_to_process(run_id, run_manager, websocket)
    )

    # Wait for either task to finish
    _, websocket_manager.pending_tasks[websocket] = await asyncio.wait(
        [output_task, input_task],
        return_when=asyncio.FIRST_COMPLETED,
    )
    websocket_manager.disconnect(websocket)



from fastapi import Query
from pydantic import BaseModel


class Pagination(BaseModel):
    page: int = Query(1, gt=0)
    limit: int = Query(10**5, gt=0)

    def offset(self) -> int:
        return (self.page - 1) * self.limit



"""
JSON Converter
---------------

Converts a user project's frontend graph to a script understandable by DFF json-importer.
"""
from pathlib import Path
from typing import List, Optional, Tuple

from omegaconf.dictconfig import DictConfig

from chatsky_ui.api.deps import get_index
from chatsky_ui.core.logger_config import get_logger
from chatsky_ui.core.config import settings
from chatsky_ui.db.base import read_conf, write_conf
from chatsky_ui.services.index import Index

logger = get_logger(__name__)


def _get_db_paths(build_id: int) -> Tuple[Path, Path, Path, Path]:
    """Get paths to frontend graph, dff script, and dff custom conditions files."""
    frontend_graph_path = settings.frontend_flows_path
    custom_conditions_file = settings.conditions_path
    custom_responses_file = settings.responses_path
    script_path = settings.scripts_dir / f"build_{build_id}.yaml"

    if not frontend_graph_path.exists():
        raise FileNotFoundError(f"File {frontend_graph_path} doesn't exist")
    if not custom_conditions_file.exists():
        raise FileNotFoundError(f"File {custom_conditions_file} doesn't exist")
    if not custom_responses_file.exists():
        raise FileNotFoundError(f"File {custom_responses_file} doesn't exist")
    if not script_path.exists():
        script_path.parent.mkdir(parents=True, exist_ok=True)
        script_path.touch()

    return frontend_graph_path, script_path, custom_conditions_file, custom_responses_file


def _organize_graph_according_to_nodes(flow_graph: DictConfig, script: dict) -> dict:
    nodes = {}
    for flow in flow_graph["flows"]:
        node_names_in_one_flow = []
        for node in flow.data.nodes:
            if "flags" in node.data:
                if "start" in node.data.flags:
                    if "start_label" in script["CONFIG"]:
                        raise ValueError("There are more than one start node in the script")
                    script["CONFIG"]["start_label"] = [flow.name, node.data.name]
                if "fallback" in node.data.flags:
                    if "fallback_label" in script["CONFIG"]:
                        raise ValueError("There are more than one fallback node in the script")
                    script["CONFIG"]["fallback_label"] = [flow.name, node.data.name]

            if node.data.name in node_names_in_one_flow:
                raise ValueError(f"There is more than one node with the name '{node.data.name}' in the same flow.")
            node_names_in_one_flow.append(node.data.name)
            nodes[node.id] = {"info": node}
            nodes[node.id]["flow"] = flow.name
            nodes[node.id]["TRANSITIONS"] = []
    return nodes


def _get_condition(nodes: dict, edge: DictConfig) -> Optional[DictConfig]:
    """Get node's condition from `nodes` according to `edge` info."""
    return next(
        (condition for condition in nodes[edge.source]["info"].data.conditions if condition["id"] == edge.sourceHandle),
        None,
    )


def _write_list_to_file(conditions_lines: list, custom_conditions_file: Path) -> None:
    """Write dff custom conditions from list to file."""
    # TODO: make reading and writing conditions async
    with open(custom_conditions_file, "w", encoding="UTF-8") as file:
        for line in conditions_lines:
            if not line.endswith("\n"):
                line = "".join([line, "\n"])
            file.write(line)


def _add_transitions(nodes: dict, edge: DictConfig, condition: DictConfig) -> None:
    """Add transitions to a node according to `edge` and `condition`."""
    # if the edge is a link_node, we add transition of its source and target
    if nodes[edge.target]["info"].type == "link_node":
        flow = nodes[edge.target]["info"].data.transition.target_flow

        target_node = nodes[edge.target]["info"].data.transition.target_node
        node = nodes[target_node]["info"].data.name
    else:
        flow = nodes[edge.target]["flow"]
        node = nodes[edge.target]["info"].data.name
    nodes[edge.source]["TRANSITIONS"].append(
        {
            "lbl": [
                flow,
                node,
                condition.data.priority,
            ],
            "cnd": f"custom_dir.conditions.{condition.name}",
        }
    )


def _fill_nodes_into_script(nodes: dict, script: dict) -> None:
    """Fill nodes into dff script dictunary."""
    for _, node in nodes.items():
        if node["info"].type == "link_node":
            continue
        if node["flow"] not in script:
            script[node["flow"]] = {}
        script[node["flow"]].update(
            {
                node["info"].data.name: {
                    "RESPONSE": node["info"].data.response,
                    "TRANSITIONS": node["TRANSITIONS"],
                }
            }
        )


def _append(service: DictConfig, services_lines: list) -> list:
    """Append a condition to a list"""
    if service.type == "python":
        service_with_newline = "".join([service.data.python.action + "\n\n"])

    logger.debug("Service to append: %s", service_with_newline)
    logger.debug("services_lines before appending: %s", services_lines)

    all_lines = services_lines + service_with_newline.split("\n")
    return all_lines


async def _shift_cnds_in_index(index: Index, cnd_strt_lineno: int, diff_in_lines: int) -> None:
    """Update the start line number of conditions in index by shifting them by `diff_in_lines`."""
    services = index.get_services()
    for _, service in services.items():
        if service["type"] == "condition":
            if service["lineno"] - 1 > cnd_strt_lineno:  # -1 is here to convert from file numeration to list numeration
                service["lineno"] += diff_in_lines

    await index.indexit_all(
        [service_name for service_name, _ in services.items()],
        [service["type"] for _, service in services.items()],
        [service["lineno"] for _, service in services.items()],
    )


async def _replace(service: DictConfig, services_lines: list, cnd_strt_lineno: int, index: Index) -> list:
    """Replace a servuce in a services list with a new one.

    Args:
        service: service to replace. `condition.data.python.action` is a string with the new service(condition)
        conditions_lines: list of conditions lines
        cnd_strt_lineno: a pointer to the service start line in custom conditions file
        index: index object to update

    Returns:
        list of all conditions as lines
    """
    cnd_strt_lineno = cnd_strt_lineno - 1  # conversion from file numeration to list numeration
    all_lines = services_lines.copy()
    if service.type == "python":
        condition = "".join([service.data.python.action + "\n\n"])
    new_cnd_lines = condition.split("\n")

    old_cnd_lines_num = 0
    for lineno, line in enumerate(all_lines[cnd_strt_lineno:]):
        if line.startswith("def ") and lineno != 0:
            break
        old_cnd_lines_num += 1

    next_func_location = cnd_strt_lineno + old_cnd_lines_num

    logger.debug("new_cnd_lines\n")
    logger.debug(new_cnd_lines)
    all_lines = all_lines[:cnd_strt_lineno] + new_cnd_lines + all_lines[next_func_location:]

    diff_in_lines = len(new_cnd_lines) - old_cnd_lines_num
    logger.debug("diff_in_lines: %s", diff_in_lines)
    logger.debug("cnd_strt_lineno: %s", cnd_strt_lineno)

    await _shift_cnds_in_index(index, cnd_strt_lineno, diff_in_lines)
    return all_lines


async def update_responses_lines(nodes: dict, responses_lines: list, index: Index) -> Tuple[dict, List[str]]:
    """Organizes the responses in nodes in a format that json-importer accepts.

    If the response type is "python", its function will be added to responses_lines to be written
    to the custom_conditions_file later.
    * If the response already exists in the responses_lines, it will be replaced with the new one.
    """
    for node in nodes.values():
        if node["info"].type == "link_node":
            continue
        response = node["info"].data.response
        logger.debug("response type: %s", response.type)
        if response.type == "python":
            response.data = response.data[0]
            if response.name not in (rsp_names := index.index):
                logger.debug("Adding response: %s", response.name)
                rsp_lineno = len(responses_lines)
                responses_lines = _append(response, responses_lines)
                await index.indexit(response.name, "response", rsp_lineno + 1)
            else:
                logger.debug("Replacing response: %s", response.name)
                responses_lines = await _replace(response, responses_lines, rsp_names[response.name]["lineno"], index)
            node["info"].data.response = f"custom_dir.responses.{response.name}"
        elif response.type == "text":
            response.data = response.data[0]
            logger.debug("Adding response: %s", response.data.text)
            node["info"].data.response = {"dff.Message": {"text": response.data.text}}
        elif response.type == "choice":
            # logger.debug("Adding response: %s", )
            dff_responses = []
            for message in response.data:
                if "text" in message:
                    dff_responses.append({"dff.Message": {"text": message["text"]}})
                else:
                    raise ValueError("Unknown response type. There must be a 'text' field in each message.")
            node["info"].data.response = {"dff.rsp.choice": dff_responses.copy()}
        else:
            raise ValueError(f"Unknown response type: {response.type}")
    return nodes, responses_lines


async def converter(build_id: int) -> None:
    """Translate frontend flow script into dff script."""
    index = get_index()
    await index.load()
    index.logger.debug("Loaded index '%s'", index.index)

    frontend_graph_path, script_path, custom_conditions_file, custom_responses_file = _get_db_paths(build_id)

    script = {
        "CONFIG": {"custom_dir": str("/" / settings.custom_dir)},
    }
    flow_graph: DictConfig = await read_conf(frontend_graph_path)  # type: ignore

    nodes = _organize_graph_according_to_nodes(flow_graph, script)

    with open(custom_responses_file, "r", encoding="UTF-8") as file:
        responses_lines = file.readlines()

    nodes, responses_lines = await update_responses_lines(nodes, responses_lines, index)

    with open(custom_conditions_file, "r", encoding="UTF-8") as file:
        conditions_lines = file.readlines()

    for flow in flow_graph["flows"]:
        for edge in flow.data.edges:
            if edge.source in nodes and edge.target in nodes:
                condition = _get_condition(nodes, edge)
                if condition is None:
                    logger.error(
                        "A condition of edge '%s' - '%s' and id of '%s' is not found in the corresponding node",
                        edge.source,
                        edge.target,
                        edge.sourceHandle,
                    )
                    continue

                if condition.name not in (cnd_names := index.index):
                    logger.debug("Adding condition: %s", condition.name)
                    cnd_lineno = len(conditions_lines)
                    conditions_lines = _append(condition, conditions_lines)
                    await index.indexit(condition.name, "condition", cnd_lineno + 1)
                else:
                    logger.debug("Replacing condition: %s", condition.name)
                    conditions_lines = await _replace(
                        condition, conditions_lines, cnd_names[condition.name]["lineno"], index
                    )

                _add_transitions(nodes, edge, condition)
            else:
                logger.error("A node of edge '%s-%s' is not found in nodes", edge.source, edge.target)

    _fill_nodes_into_script(nodes, script)

    _write_list_to_file(conditions_lines, custom_conditions_file)
    _write_list_to_file(responses_lines, custom_responses_file)
    await write_conf(script, script_path)



"""
Index service
-------------

The Index service is responsible for indexing the user bot's conditions, responses, and services.
By indexing the project, the Index service creates an in-memory representation that can be
quickly accessed when needed.
"""
import asyncio
from pathlib import Path
from typing import Dict, List, Optional

from omegaconf import OmegaConf
from omegaconf.dictconfig import DictConfig

from chatsky_ui.core.config import settings
from chatsky_ui.core.logger_config import get_logger
from chatsky_ui.db.base import read_conf, read_logs, write_conf


class Index:
    def __init__(self):
        self.index: dict = {}
        self.conditions: List[str] = []
        self.responses: List[str] = []
        self.services: List[str] = []
        self._logger = None
        self._path = None

    @property
    def logger(self):
        if self._logger is None:
            raise ValueError("Logger has not been configured. Call set_logger() first.")
        return self._logger

    def set_logger(self):
        self._logger = get_logger(__name__)

    @property
    def path(self):
        if self._path is None:
            raise ValueError("Path has not been configured. Call set_path() first.")
        return self._path

    def set_path(self, path: Path):
        self._path = path

    async def _load_index(self) -> None:
        """Load indexed conditions, responses and services from disk."""
        db_index: DictConfig = await read_conf(self.path)  # type: ignore
        index_dict: Dict[str, dict] = OmegaConf.to_container(db_index, resolve=True)  # type: ignore
        self.index = index_dict
        self.logger.debug("Index loaded")

    async def _load_conditions(self) -> None:
        """Load conditions from disk."""
        path = settings.conditions_path
        if path.exists():
            self.conditions = await read_logs(path)
            self.logger.debug("Conditions loaded")
        else:
            self.logger.warning("No conditions file found")

    async def _load_responses(self) -> None:
        """Load responses from disk."""
        path = settings.responses_path
        if path.exists():
            self.responses = await read_logs(path)
            self.logger.debug("Responses loaded")
        else:
            self.logger.warning("No responses file found")

    async def _load_services(self) -> None:
        """Load services from disk."""
        path = settings.responses_path.parent / "services.py"
        if path.exists():
            self.services = await read_logs(path)
            self.logger.debug("Services loaded")
        else:
            self.logger.warning("No services file found")

    def _get_service_code(self, services_lst: List[str], lineno: int) -> List[str]:
        """Get service code from services list.

        Example:
            >>> _get_service_code(["def is_upper_case(name):\n", "    return name.isupper()"], 1)
            ['def is_upper_case(name):\n', '    return name.isupper()']
        """
        service: List[str] = []
        func_lines: List[str] = services_lst[lineno - 1 :]
        self.logger.debug("services_lst: %s", services_lst)
        for func_lineno, func_line in enumerate(func_lines):
            if func_line.startswith("def ") and func_lineno != 0:
                break
            service.append(func_line)  # ?maybe with \n
        return service

    async def load(self) -> None:
        """Load index and services into memory."""
        if not self.path.exists():
            raise FileNotFoundError(f"File {self.path} doesn't exist")

        await asyncio.gather(
            self._load_index(),
            self._load_conditions(),
            self._load_responses(),
            self._load_services(),
        )
        self.logger.info("Index and services loaded")
        self.logger.debug("Loaded index: %s", self.index)

    def get_services(self) -> dict:
        """Get indexed services.

        Example:
            >>> get_services()
            {
                "is_upper_case": {"type": "condition", "lineno": 3},
                "say_hi": {"type": "response", "lineno": 5}
            }
        """
        return self.index

    async def search_service(self, service_name: str) -> Optional[List[str]]:
        """Get the body code of a service based on its indexed info (type, lineno).

        Example:
            >>> search_service("is_upper_case")
            ["def is_upper_case(name):\n", "    return name.isupper()"]

        """
        if service_name not in self.index:
            return []
        service_type = self.index[service_name]["type"]
        lineno = int(self.index[service_name]["lineno"])

        if service_type == "condition":
            return self._get_service_code(self.conditions, lineno)
        elif service_type == "response":
            return self._get_service_code(self.responses, lineno)
        elif service_type == "service":
            return self._get_service_code(self.services, lineno)

    async def indexit(self, service_name: str, type_: str, lineno: int) -> None:
        """Add service info to the index using indexit_all method."""
        self.logger.debug("Indexing '%s'", service_name)
        await self.indexit_all([service_name], [type_], [lineno])
        self.logger.info("Indexed '%s'", service_name)

    async def indexit_all(self, services_names: List[str], types: List[str], linenos: List[int]) -> None:
        """Index multiple services.

        The index is added to the index in the form: {service_name: {"type": ``type_``, "lineno": lineno}}.

        Args:
            services_names: list of service names
            types: list of service types ("condition", "response", "service")
            linenos: list of service starting line numbers according to its place in the file.

        Raises:
            FileNotFoundError: if the index file doesn't exist

        Example:
            >>> services_names = ["is_upper_case", "say_hi"]
            >>> types = ["condition", "response"]
            >>> linenos = [3, 5]
            >>> await indexit_all(services_names, types, linenos)
            {
                "is_upper_case": {"type": "condition", "lineno": 3},
                "say_hi": {"type": "response", "lineno": 5}
            }

        Returns:
            None
        """
        if not self.path.exists():
            raise FileNotFoundError(f"File {self.path} doesn't exist")

        for service_name, type_, lineno in zip(services_names, types, linenos):
            self.index[service_name] = {
                "type": type_,  # condition/response/service
                "lineno": lineno,
            }

        await write_conf(self.index, self.path)  # ?to background tasks



import os
from pathlib import Path

import uvicorn
from dotenv import load_dotenv
from omegaconf import DictConfig, OmegaConf

load_dotenv()


class Settings:
    def __init__(self):
        self.API_V1_STR = "/api/v1"
        self.APP = "chatsky_ui.main:app"

        self.config_file_path = Path(__file__).absolute()
        self.static_files = self.config_file_path.parent.with_name("static")
        self.start_page = self.static_files / "index.html"
        self.package_dir = self.config_file_path.parents[2]
        self.temp_conf = self.config_file_path.with_name("temp_conf.yaml")

        self.set_config(
            host=os.getenv("HOST", "0.0.0.0"),
            port=os.getenv("PORT", "8000"),
            log_level=os.getenv("LOG_LEVEL", "info"),
            conf_reload=os.getenv("CONF_RELOAD", "false"),
            work_directory=".",
        )

    def set_config(self, **kwargs):
        for key, value in kwargs.items():
            if key == "work_directory":
                value = Path(value)
            elif key == "conf_reload":
                value = str(value).lower() in ["true", "yes", "t", "y", "1"]
            elif key == "port":
                value = int(value)
            setattr(self, key, value)

        if "work_directory" in kwargs:
            self._set_user_proj_paths()

    def _set_user_proj_paths(self):
        self.builds_path = self.work_directory / "chatsky_ui/app_data/builds.yaml"
        self.runs_path = self.work_directory / "chatsky_ui/app_data/runs.yaml"
        self.frontend_flows_path = self.work_directory / "chatsky_ui/app_data/frontend_flows.yaml"
        self.dir_logs = self.work_directory / "chatsky_ui/logs"
        self.presets = self.work_directory / "chatsky_ui/presets"
        self.snippet2lint_path = self.work_directory / "chatsky_ui/.snippet2lint.py"

        self.custom_dir = self.work_directory / "bot/custom"
        self.index_path = self.custom_dir / ".services_index.yaml"
        self.conditions_path = self.custom_dir / "conditions.py"
        self.responses_path = self.custom_dir / "responses.py"
        self.scripts_dir = self.work_directory / "bot/scripts"

    def save_config(self):
        if not self.temp_conf.exists():
            self.temp_conf.touch()
        OmegaConf.save(
            OmegaConf.create(
                {
                    "work_directory": str(self.work_directory),
                    "host": self.host,
                    "port": self.port,
                    "log_level": self.log_level,
                    "conf_reload": self.conf_reload,
                }
            ),  # type: ignore
            self.temp_conf,
        )

    def _load_temp_config(self) -> DictConfig:
        if not self.temp_conf.exists():
            raise FileNotFoundError(f"{self.temp_conf} not found.")

        return OmegaConf.load(self.temp_conf)  # type: ignore

    def refresh_work_dir(self):
        config = self._load_temp_config()
        self.set_config(**config)


class AppRunner:
    def __init__(self):
        self._settings = None

    @property
    def settings(self):
        if self._settings is None:
            raise ValueError("Settings has not been configured. Call set_logger() first.")
        return self._settings

    def set_settings(self, app_settings: Settings):
        self._settings = app_settings

    def run(self):
        if reload := self.settings.conf_reload:
            reload_conf = {
                "reload": reload,
                "reload_dirs": [str(self.settings.package_dir)],
                "reload_excludes": [
                    f"./{self.settings.work_directory}/*",
                    f"./{self.settings.work_directory}/*/*",
                    f"./{self.settings.work_directory}/*/*/*",
                ],
            }
        else:
            reload_conf = {"reload": reload}
        uvicorn.run(
            self.settings.APP,
            host=self.settings.host,
            port=self.settings.port,
            log_level=self.settings.log_level,
            **reload_conf,
        )


settings = Settings()
app_runner = AppRunner()



from chatsky_ui.core.config import settings
from chatsky_ui.services.index import Index
from chatsky_ui.services.process_manager import BuildManager, RunManager
from chatsky_ui.services.websocket_manager import WebSocketManager

build_manager = BuildManager()


def get_build_manager() -> BuildManager:
    build_manager.set_logger()
    return build_manager


run_manager = RunManager()


def get_run_manager() -> RunManager:
    run_manager.set_logger()
    return run_manager


websocket_manager = WebSocketManager()


def get_websocket_manager() -> WebSocketManager:
    websocket_manager.set_logger()
    return websocket_manager


index = Index()


def get_index() -> Index:
    index.set_logger()
    index.set_path(settings.index_path)
    return index



import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, Literal, Optional

from chatsky_ui.core.config import settings

LOG_LEVELS: Dict[str, int] = {
    "critical": logging.CRITICAL,
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
}


def setup_logging(log_type: Literal["builds", "runs"], id_: int, timestamp: datetime) -> Path:
    # Ensure log_type is either 'builds' or 'runs'
    if log_type not in ["builds", "runs"]:
        raise ValueError("log_type must be 'builds' or 'runs'")

    # Get today's date separated with '_' using the timestamp
    log_name = "_".join([str(id_), timestamp.strftime("%H:%M:%S")])
    today_date = timestamp.strftime("%Y%m%d")

    log_directory = settings.dir_logs / log_type / today_date

    log_directory.mkdir(parents=True, exist_ok=True)

    log_file = log_directory / f"{log_name}.log"
    log_file.touch(exist_ok=True)
    return log_file


def get_logger(name, file_handler_path: Optional[Path] = None):
    if file_handler_path is None:
        file_handler_path = settings.dir_logs / "logs.log"
    if not file_handler_path.parent.exists():
        raise FileNotFoundError(f"File {file_handler_path} doesn't exist")
    file_handler_path.touch(exist_ok=True)

    logger = logging.getLogger(name)
    logger.propagate = False
    logger.setLevel(LOG_LEVELS[settings.log_level])

    if not logger.hasHandlers():
        c_handler = logging.StreamHandler()
        f_handler = logging.FileHandler(file_handler_path)
        c_handler.setLevel(LOG_LEVELS[settings.log_level])
        f_handler.setLevel(LOG_LEVELS[settings.log_level])

        c_format = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
        f_format = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        c_handler.setFormatter(c_format)
        f_handler.setFormatter(f_format)

        logger.addHandler(c_handler)
        logger.addHandler(f_handler)

    return logger



from contextlib import asynccontextmanager

from fastapi import APIRouter, FastAPI, Response
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, HTMLResponse, RedirectResponse

from chatsky_ui.api.api_v1.api import api_router
from chatsky_ui.api.deps import get_index
from chatsky_ui.core.config import settings

index_dict = {}


@asynccontextmanager
async def lifespan(app: FastAPI):
    if settings.temp_conf.exists():
        settings.refresh_work_dir()

    index_dict["instance"] = get_index()
    await index_dict["instance"].load()
    yield

    settings.temp_conf.unlink(missing_ok=True)


app = FastAPI(title="DF Designer", lifespan=lifespan)


app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

root_router = APIRouter()


@root_router.get("/app/{path:path}")
async def route_static_file(path: str):
    if not settings.start_page.exists():
        return HTMLResponse(content="frontend is not built")
    file_path = settings.static_files / path.split("/")[-1]
    if file_path.suffix in (".js", ".css", ".html", ".ttf"):
        return FileResponse(file_path)
    return FileResponse(settings.static_files / "index.html")


@root_router.get("/")
async def root() -> Response:
    """Redirect '/' to index.html"""
    return RedirectResponse(url="/app")


app.include_router(root_router)
app.include_router(api_router)



from fastapi import APIRouter

from chatsky_ui.api.api_v1.endpoints import bot, config, dff_services, flows
from chatsky_ui.core.config import settings

api_router = APIRouter()

api_router.include_router(config.router, prefix="/".join([settings.API_V1_STR, "config"]), tags=["config"])
api_router.include_router(flows.router, prefix="/".join([settings.API_V1_STR, "flows"]), tags=["flows"])
api_router.include_router(dff_services.router, prefix="/".join([settings.API_V1_STR, "services"]), tags=["services"])
api_router.include_router(bot.router, prefix="/".join([settings.API_V1_STR, "bot"]), tags=["bot"])



# pylint: disable=C0413
# flake8: noqa: E402

from contextlib import asynccontextmanager
from typing import Generator

import httpx
import nest_asyncio
import pytest
from fastapi.testclient import TestClient
from httpx import AsyncClient

nest_asyncio.apply = lambda: None

from chatsky_ui.main import app
from chatsky_ui.schemas.pagination import Pagination
from chatsky_ui.schemas.preset import Preset
from chatsky_ui.services.process import RunProcess
from chatsky_ui.services.process_manager import BuildManager, RunManager
from chatsky_ui.services.websocket_manager import WebSocketManager

DUMMY_BUILD_ID = -1


async def start_process(async_client: AsyncClient, endpoint, preset_end_status) -> httpx.Response:
    return await async_client.post(
        endpoint,
        json={"wait_time": 0.1, "end_status": preset_end_status},
    )


@asynccontextmanager
async def override_dependency(mocker_obj, get_manager_func):
    process_manager = get_manager_func()
    process_manager.check_status = mocker_obj.AsyncMock()
    app.dependency_overrides[get_manager_func] = lambda: process_manager
    try:
        yield process_manager
    finally:
        for _, process in process_manager.processes.items():
            if process.process.returncode is None:
                await process.stop()
        app.dependency_overrides = {}


@pytest.fixture
def client() -> Generator:
    with TestClient(app=app) as client:
        yield client


@pytest.fixture(scope="session")
def preset() -> Preset:
    return Preset(
        wait_time=0,
        end_status="loop",
    )


@pytest.fixture
def pagination() -> Pagination:
    return Pagination()


@pytest.fixture()
def run_process():
    async def _run_process(cmd_to_run):
        process = RunProcess(id_=0, build_id=DUMMY_BUILD_ID)
        await process.start(cmd_to_run)
        return process

    return _run_process


@pytest.fixture()
def run_manager():
    manager = RunManager()
    manager.set_logger()
    return manager


@pytest.fixture()
def build_manager():
    return BuildManager()


@pytest.fixture
def websocket_manager():
    manager = WebSocketManager()
    manager.set_logger()
    return manager



from typing import Dict, Union

from fastapi import APIRouter
from omegaconf import OmegaConf

from chatsky_ui.core.config import settings
from chatsky_ui.db.base import read_conf, write_conf

router = APIRouter()


@router.get("/")
async def flows_get() -> Dict[str, Union[str, Dict[str, list]]]:
    """Get the flows by reading the frontend_flows.yaml file."""
    omega_flows = await read_conf(settings.frontend_flows_path)
    dict_flows = OmegaConf.to_container(omega_flows, resolve=True)
    return {"status": "ok", "data": dict_flows}  # type: ignore


@router.post("/")
async def flows_post(flows: Dict[str, list]) -> Dict[str, str]:
    """Write the flows to the frontend_flows.yaml file."""
    await write_conf(flows, settings.frontend_flows_path)
    return {"status": "ok"}



from asyncio import Lock
from pathlib import Path
from typing import List, Union

import aiofiles
from omegaconf import OmegaConf
from omegaconf.dictconfig import DictConfig
from omegaconf.listconfig import ListConfig

file_lock = Lock()


async def read_conf(path: Path) -> Union[DictConfig, ListConfig]:
    async with file_lock:
        async with aiofiles.open(path, "r", encoding="UTF-8") as file:
            data = await file.read()
    omega_data = OmegaConf.create(data)  # read from a YAML string
    return omega_data


async def write_conf(data: Union[DictConfig, ListConfig, dict, list], path: Path) -> None:
    yaml_conf = OmegaConf.to_yaml(data)
    async with file_lock:
        async with aiofiles.open(path, "w", encoding="UTF-8") as file:  # TODO: change to "a" for append
            await file.write(yaml_conf)


async def read_logs(log_file: Path) -> List[str]:
    logs = []
    if not log_file.exists():
        raise FileNotFoundError(f"Log file '{log_file}' not found")
    async with aiofiles.open(log_file, "r", encoding="UTF-8") as file:
        logs = [line async for line in file]
    return logs



import ast


def get_imports_from_file(file_path):
    """Gets all imports from a file and returns them as strings separated by newlines.

    Example:
    >>> get_imports_from_file('file.py')
    'import os\nfrom pathlib import Path'
    """
    with open(file_path, "r", encoding="UTF-8") as file:
        node = ast.parse(file.read(), file_path)

    imports = []
    for item in ast.walk(node):
        if isinstance(item, ast.Import):
            for alias in item.names:
                if alias.asname:
                    imports.append(f"import {alias.name} as {alias.asname}")
                else:
                    imports.append(f"import {alias.name}")
        elif isinstance(item, ast.ImportFrom):
            module = item.module if item.module else ""
            for alias in item.names:
                if alias.asname:
                    imports.append(f"from {module} import {alias.name} as {alias.asname}")
                else:
                    imports.append(f"from {module} import {alias.name}")

    return "\n".join(imports)



import pytest
from fastapi import BackgroundTasks, HTTPException

from chatsky_ui.api.api_v1.endpoints.bot import (
    _check_process_status,
    _stop_process,
    check_build_processes,
    check_run_processes,
    connect,
    get_build_logs,
    get_run_logs,
    start_build,
    start_run,
)
from chatsky_ui.schemas.process_status import Status
from chatsky_ui.services.process_manager import RunManager

PROCESS_ID = 0
RUN_ID = 42
BUILD_ID = 43


@pytest.mark.asyncio
async def test_stop_process_success(mocker):
    process_manager = mocker.MagicMock()
    process_manager.stop = mocker.AsyncMock()

    # Call the function under test
    await _stop_process(PROCESS_ID, process_manager)

    # Assert the stop method was called once with the correct id
    process_manager.stop.assert_awaited_once_with(PROCESS_ID)


# TODO: take into consideration the errors when process type is build
@pytest.mark.parametrize("error_type", [RuntimeError, ProcessLookupError])
@pytest.mark.asyncio
async def test_stop_process_error(mocker, error_type):
    mock_stop = mocker.AsyncMock(side_effect=error_type)
    mocker.patch.object(RunManager, "stop", mock_stop)

    process_type = "run"

    with pytest.raises(HTTPException) as exc_info:
        await _stop_process(PROCESS_ID, RunManager(), process_type)

        # Assert the stop method was called once with the correct id
        assert exc_info.value.status_code == 404
    mock_stop.assert_awaited_once_with(PROCESS_ID)


# TODO: check the errors
@pytest.mark.asyncio
async def test_check_process_status(mocker):
    mocked_process_manager = mocker.MagicMock()
    mocker.patch.object(mocked_process_manager, "processes", {PROCESS_ID: mocker.MagicMock()})
    mocker.patch.object(mocked_process_manager, "get_status", mocker.AsyncMock(return_value=Status.ALIVE))

    response = await _check_process_status(PROCESS_ID, mocked_process_manager)

    assert response == {"status": "alive"}
    mocked_process_manager.get_status.assert_awaited_once_with(0)


@pytest.mark.asyncio
async def test_start_build(mocker):
    build_manager = mocker.MagicMock()
    preset = mocker.MagicMock()

    start = mocker.AsyncMock(return_value=BUILD_ID)
    mocker.patch.multiple(build_manager, start=start, check_status=mocker.AsyncMock())
    mocker.patch.multiple(preset, wait_time=0, end_status="loop")

    response = await start_build(preset, background_tasks=BackgroundTasks(), build_manager=build_manager)
    start.assert_awaited_once_with(preset)
    assert response == {"status": "ok", "build_id": BUILD_ID}


@pytest.mark.asyncio
async def test_check_build_processes_some_info(mocker, pagination):
    build_manager = mocker.AsyncMock()
    run_manager = mocker.AsyncMock()

    await check_build_processes(BUILD_ID, build_manager, run_manager, pagination)

    build_manager.get_build_info.assert_awaited_once_with(BUILD_ID, run_manager)


@pytest.mark.asyncio
async def test_check_build_processes_all_info(mocker, pagination):
    build_id = None
    build_manager = mocker.AsyncMock()
    run_manager = mocker.AsyncMock()

    await check_build_processes(build_id, build_manager, run_manager, pagination)

    build_manager.get_full_info_with_runs_info.assert_awaited_once_with(
        run_manager, offset=pagination.offset(), limit=pagination.limit
    )


@pytest.mark.asyncio
async def test_get_build_logs(mocker, pagination):
    build_manager = mocker.AsyncMock()

    await get_build_logs(BUILD_ID, build_manager, pagination)

    build_manager.fetch_build_logs.assert_awaited_once_with(BUILD_ID, pagination.offset(), pagination.limit)


@pytest.mark.asyncio
async def test_start_run(mocker):
    run_manager = mocker.MagicMock()
    preset = mocker.MagicMock()

    start = mocker.AsyncMock(return_value=RUN_ID)
    mocker.patch.multiple(run_manager, start=start, check_status=mocker.AsyncMock())
    mocker.patch.multiple(preset, wait_time=0, end_status="loop")

    response = await start_run(
        build_id=BUILD_ID, preset=preset, background_tasks=BackgroundTasks(), run_manager=run_manager
    )
    start.assert_awaited_once_with(BUILD_ID, preset)
    assert response == {"status": "ok", "run_id": RUN_ID}


@pytest.mark.asyncio
async def test_check_run_processes_some_info(mocker, pagination):
    run_manager = mocker.AsyncMock()

    await check_run_processes(RUN_ID, run_manager, pagination)

    run_manager.get_run_info.assert_awaited_once_with(RUN_ID)


@pytest.mark.asyncio
async def test_check_run_processes_all_info(mocker, pagination):
    run_id = None
    run_manager = mocker.AsyncMock()

    await check_run_processes(run_id, run_manager, pagination)

    run_manager.get_full_info.assert_awaited_once_with(offset=pagination.offset(), limit=pagination.limit)


@pytest.mark.asyncio
async def test_get_run_logs(mocker, pagination):
    run_manager = mocker.AsyncMock()

    await get_run_logs(RUN_ID, run_manager, pagination)

    run_manager.fetch_run_logs.assert_awaited_once_with(RUN_ID, pagination.offset(), pagination.limit)


@pytest.mark.asyncio
async def test_connect(mocker):
    websocket = mocker.AsyncMock()
    websocket_manager = mocker.AsyncMock()
    websocket_manager.disconnect = mocker.MagicMock()
    run_manager = mocker.MagicMock()
    run_process = mocker.MagicMock()
    run_manager.processes = {RUN_ID: run_process}
    mocker.patch.object(websocket, "query_params", {"run_id": str(RUN_ID)})

    await connect(websocket, websocket_manager, run_manager)

    websocket_manager.connect.assert_awaited_once_with(websocket)
    websocket_manager.send_process_output_to_websocket.assert_awaited_once_with(RUN_ID, run_manager, websocket)
    websocket_manager.forward_websocket_messages_to_process.assert_awaited_once_with(RUN_ID, run_manager, websocket)
    websocket_manager.disconnect.assert_called_once_with(websocket)



import asyncio

import httpx
import pytest
from httpx_ws import aconnect_ws
from httpx_ws.transport import ASGIWebSocketTransport

from chatsky_ui.api.deps import get_build_manager, get_run_manager
from chatsky_ui.main import app
from chatsky_ui.schemas.process_status import Status
from chatsky_ui.tests.conftest import override_dependency, start_process


async def _assert_process_status(response, process_manager):
    assert response.json().get("status") == "ok", "Start process response status is not 'ok'"
    process_manager.check_status.assert_awaited_once()


@pytest.mark.asyncio
async def test_all(mocker):
    async with httpx.AsyncClient(transport=ASGIWebSocketTransport(app)) as client:
        async with override_dependency(mocker, get_build_manager) as process_manager:
            response = await start_process(
                client,
                endpoint="http://localhost:8000/api/v1/bot/build/start",
                preset_end_status="success",
            )

            build_id = process_manager.get_last_id()

            await _assert_process_status(response, process_manager)
            try:
                await asyncio.wait_for(process_manager.processes[build_id].process.wait(), timeout=20)
            except asyncio.exceptions.TimeoutError as exc:
                raise Exception(
                    "Process with expected end status Status.ALIVE timed out with status Status.RUNNING."
                ) from exc
            assert await process_manager.get_status(build_id) == Status.COMPLETED

        async with override_dependency(mocker, get_run_manager) as process_manager:
            response = await start_process(
                client,
                endpoint=f"http://localhost:8000/api/v1/bot/run/start/{build_id}",
                preset_end_status="success",
            )

            run_id = process_manager.get_last_id()

            await _assert_process_status(response, process_manager)
            await asyncio.sleep(10)
            assert await process_manager.get_status(run_id) == Status.ALIVE

            async with aconnect_ws(f"http://localhost:8000/api/v1/bot/run/connect?run_id={run_id}", client) as ws:
                message = await ws.receive_text()
                assert message == "Start chatting"



from pathlib import Path

import pytest
from omegaconf import OmegaConf

RUN_ID = 42
BUILD_ID = 43


class TestRunManager:
    @pytest.mark.asyncio
    async def test_start(self, mocker, preset, run_manager):  # noqa: F811
        # Mock the RunProcess constructor whereever it's called in
        # the process_manager file within the scope of this test function
        run_process = mocker.patch("chatsky_ui.services.process_manager.RunProcess")
        run_process_instance = run_process.return_value
        run_process_instance.start = mocker.AsyncMock()
        run_manager.get_full_info = mocker.AsyncMock(return_value=[{"id": RUN_ID}])

        await run_manager.start(build_id=BUILD_ID, preset=preset)

        run_process.assert_called_once_with(run_manager.last_id, BUILD_ID, preset.end_status)
        run_process_instance.start.assert_awaited_once_with(
            f"chatsky.ui run_bot --build-id {BUILD_ID} " f"--preset {preset.end_status} " f"--project-dir ."
        )

        assert run_manager.processes[run_manager.last_id] is run_process_instance

    @pytest.mark.asyncio
    async def test_stop_success(self, mocker, run_manager):
        run_manager.processes[RUN_ID] = mocker.MagicMock()
        run_manager.processes[RUN_ID].stop = mocker.AsyncMock()

        await run_manager.stop(RUN_ID)
        run_manager.processes[RUN_ID].stop.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test_stop_with_error(self, run_manager):
        with pytest.raises((RuntimeError, ProcessLookupError)):
            await run_manager.stop(RUN_ID)

    # def test_check_status(self, run_manager, preset):
    # pass

    @pytest.mark.asyncio
    async def test_get_process_info(self, mocker, run_manager):
        df_conf = OmegaConf.create(
            f"""
            - id: {RUN_ID}
              status: stopped
        """
        )
        df_conf_dict = {
            "id": RUN_ID,
            "status": "stopped",
        }

        read_conf = mocker.patch("chatsky_ui.services.process_manager.read_conf")
        read_conf.return_value = df_conf

        run_info = await run_manager.get_run_info(RUN_ID)
        assert run_info == df_conf_dict

    @pytest.mark.asyncio
    async def test_get_full_info(self, mocker, run_manager):
        df_conf = OmegaConf.create(
            f"""
            - id: {RUN_ID}
              status: stopped
            - id: {RUN_ID + 1}
              status: stopped
        """
        )
        df_conf_dict = {
            "id": RUN_ID,
            "status": "stopped",
        }

        read_conf = mocker.patch("chatsky_ui.services.process_manager.read_conf")
        read_conf.return_value = df_conf

        run_info = await run_manager.get_full_info(0, 1)
        assert run_info == [df_conf_dict]

    @pytest.mark.asyncio
    async def test_fetch_run_logs(self, mocker, run_manager):
        LOG_PATH = Path("path/to/log")
        run_manager.get_process_info = mocker.AsyncMock(return_value={"id": RUN_ID, "log_path": LOG_PATH})

        read_logs = mocker.patch("chatsky_ui.services.process_manager.read_logs", return_value=["log1", "log2"])

        logs = await run_manager.fetch_run_logs(RUN_ID, 0, 1)

        run_manager.get_process_info.assert_awaited_once()
        read_logs.assert_awaited_once_with(LOG_PATH)
        assert logs == ["log1"]



from typing import List

import dff.script.conditions as cnd
from dff.pipeline.pipeline import script_parsing

AUTO_COMPLETION_MAP = {
    "exact_match": 'cnd.exact_match(Message("hello"))(ctx, pipeline)',
    "regexp": 'cnd.regexp(r"how are you", re.IGNORECASE)(ctx, pipeline)',
    "any": 'cnd.any([hi_lower_case_condition, cnd.exact_match(Message("hello"))])(ctx, pipeline)',
    "all": 'cnd.all([cnd.regexp(r"talk"), cnd.regexp(r"about.*music")])(ctx, pipeline)',
}


def get_dff_conditions() -> List[dict]:
    """Gets the DFF's out-of-the-box conditions.

    Returns:
        List of conditions info with the following keys:
            "label": The condition name suggestions to pop up for user.
            "type": "function".
            "info": Detailed info about every condition, parsed from DFF docs.
            "apply": Autocompletion of the conditon call.
    """
    native_services = script_parsing.get_dff_objects()
    native_conditions = [k.split(".")[-1] for k, _ in native_services.items() if k.startswith("dff.cnd.")]
    cnd_full_info = []
    for condition in native_conditions:
        cnd_full_info.append(
            {
                "label": f"cnd.{condition}",
                "type": "function",
                "info": getattr(cnd, condition).__doc__,
                "apply": AUTO_COMPLETION_MAP.get(condition, "cnd.<DFF_CONDITION>()(ctx, pipeline)"),
            }
        )

    return cnd_full_info



from enum import Enum


class Status(Enum):
    NULL = "null"
    STOPPED = "stopped"
    COMPLETED = "completed"
    FAILED = "failed"
    RUNNING = "running"
    ALIVE = "alive"
    FAILED_WITH_UNEXPECTED_CODE = "failed with unexpected returncode"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chatsky_ui',
 'chatsky_ui.api',
 'chatsky_ui.api.api_v1',
 'chatsky_ui.api.api_v1.endpoints',
 'chatsky_ui.clients',
 'chatsky_ui.core',
 'chatsky_ui.db',
 'chatsky_ui.db.crud',
 'chatsky_ui.db.models',
 'chatsky_ui.schemas',
 'chatsky_ui.services',
 'chatsky_ui.tests',
 'chatsky_ui.tests.api',
 'chatsky_ui.tests.e2e',
 'chatsky_ui.tests.integration',
 'chatsky_ui.tests.services',
 'chatsky_ui.utils']

package_data = \
{'': ['*'], 'chatsky_ui': ['static/*']}

install_requires = \
['aiofiles>=23.2.1,<24.0.0',
 'cookiecutter>=2.6.0,<3.0.0',
 'dff==0.6.4.dev0',
 'fastapi>=0.110.0,<0.111.0',
 'httpx-ws>=0.6.0,<0.7.0',
 'httpx>=0.27.0,<0.28.0',
 'omegaconf>=2.3.0,<3.0.0',
 'pydantic-settings>=2.2.1,<3.0.0',
 'pydantic>=2.6.3,<3.0.0',
 'pylint>=3.2.3,<4.0.0',
 'pytest-asyncio>=0.23.6,<0.24.0',
 'pytest-mock>=3.14.0,<4.0.0',
 'pytest>=8.1.1,<9.0.0',
 'sphinx',
 'sphinx-rtd-theme',
 'typer>=0.9.0,<0.10.0',
 'uvicorn[standard]>=0.28.0,<0.29.0']

entry_points = \
{'console_scripts': ['chatsky.ui = chatsky_ui.cli:cli']}

setup_kwargs = {
    'name': 'chatsky-ui',
    'version': '0.2.0',
    'description': 'Chatsky-UI is GUI for Chatsky Framework, that is a free and open-source software stack for creating chatbots, released under the terms of Apache License 2.0.',
    'long_description': '',
    'author': 'Denis Kuznetsov',
    'author_email': 'kuznetsov.den.p@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



import pytest
from fastapi import WebSocket


class TestWebSocketManager:
    @pytest.mark.asyncio
    async def test_connect(self, mocker, websocket_manager):
        mocked_websocket = mocker.MagicMock(spec=WebSocket)

        await websocket_manager.connect(mocked_websocket)

        mocked_websocket.accept.assert_awaited_once_with()
        assert mocked_websocket in websocket_manager.active_connections

    @pytest.mark.asyncio
    async def test_disconnect(self, mocker, websocket_manager):
        mocked_websocket = mocker.MagicMock(spec=WebSocket)
        websocket_manager.active_connections.append(mocked_websocket)
        websocket_manager.pending_tasks[mocked_websocket] = set()

        websocket_manager.disconnect(mocked_websocket)

        assert mocked_websocket not in websocket_manager.pending_tasks
        assert mocked_websocket not in websocket_manager.active_connections

    @pytest.mark.asyncio
    async def test_send_process_output_to_websocket(self, mocker, websocket_manager):
        run_id = 42
        awaited_response = "Hello from DF-Designer"

        websocket = mocker.AsyncMock()
        run_manager = mocker.MagicMock()
        run_process = mocker.MagicMock()
        run_process.read_stdout = mocker.AsyncMock(side_effect=[awaited_response.encode(), None])
        run_manager.processes = {run_id: run_process}

        await websocket_manager.send_process_output_to_websocket(run_id, run_manager, websocket)

        assert run_process.read_stdout.call_count == 2
        websocket.send_text.assert_awaited_once_with(awaited_response)

    @pytest.mark.asyncio
    async def test_forward_websocket_messages_to_process(self, mocker, websocket_manager):
        run_id = 42
        awaited_message = "Hello from DF-Designer"

        websocket = mocker.AsyncMock()
        websocket.receive_text = mocker.AsyncMock(side_effect=[awaited_message, None])
        run_manager = mocker.MagicMock()
        run_process = mocker.MagicMock()
        run_process.write_stdin = mocker.AsyncMock()
        run_manager.processes = {run_id: run_process}

        await websocket_manager.forward_websocket_messages_to_process(run_id, run_manager, websocket)

        assert websocket.receive_text.await_count == 2
        run_process.write_stdin.assert_called_once_with(awaited_message.encode() + b"\n")



import asyncio
import os

import httpx
import pytest
from dotenv import load_dotenv
from fastapi import status
from httpx import ASGITransport, AsyncClient
from httpx_ws import aconnect_ws
from httpx_ws.transport import ASGIWebSocketTransport

from chatsky_ui.api.deps import get_build_manager, get_run_manager
from chatsky_ui.core.logger_config import get_logger
from chatsky_ui.main import app
from chatsky_ui.schemas.process_status import Status
from chatsky_ui.tests.conftest import override_dependency, start_process

load_dotenv()

BUILD_COMPLETION_TIMEOUT = float(os.getenv("BUILD_COMPLETION_TIMEOUT", 10))
RUN_RUNNING_TIMEOUT = float(os.getenv("RUN_RUNNING_TIMEOUT", 5))

logger = get_logger(__name__)


async def _assert_process_status(response, process_manager, expected_end_status, timeout):
    assert response.json().get("status") == "ok", "Start process response status is not 'ok'"
    process_manager.check_status.assert_awaited_once()

    try:
        await asyncio.wait_for(process_manager.processes[process_manager.last_id].process.wait(), timeout=timeout)
    except asyncio.exceptions.TimeoutError as exc:
        if expected_end_status in [Status.ALIVE, Status.RUNNING]:
            logger.debug("Loop process timed out. Expected behavior.")
        else:
            raise Exception(
                f"Process with expected end status '{expected_end_status}' timed out with status 'running'."
            ) from exc

    process_id = process_manager.last_id
    logger.debug("Process id is %s", process_id)
    current_status = await process_manager.get_status(process_id)
    assert (
        current_status == expected_end_status
    ), f"Current process status '{current_status}' did not match the expected '{expected_end_status}'"

    return current_status


async def _test_start_process(mocker_obj, get_manager_func, endpoint, preset_end_status, expected_end_status, timeout):
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as async_client:
        async with override_dependency(mocker_obj, get_manager_func) as process_manager:
            response = await start_process(async_client, endpoint, preset_end_status)
            current_status = await _assert_process_status(response, process_manager, expected_end_status, timeout)

            if current_status == Status.RUNNING:
                process = process_manager.processes[process_manager.last_id].process
                process.terminate()
                try:
                    await asyncio.wait_for(process.wait(), timeout=timeout)
                    logger.debug("The test process was gracefully terminated.")
                except asyncio.TimeoutError:
                    process.kill()
                    await process.wait()
                    logger.debug("The test process was forcefully killed.")


async def _test_stop_process(mocker, get_manager_func, start_endpoint, stop_endpoint):
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as async_client:
        async with override_dependency(mocker, get_manager_func) as manager:
            start_response = await start_process(async_client, start_endpoint, preset_end_status="loop")
            assert start_response.status_code == 201
            logger.debug("Processes: %s", manager.processes)

            last_id = manager.get_last_id()
            logger.debug("Last id: %s, type: %s", last_id, type(last_id))
            logger.debug("Process status %s", await manager.get_status(last_id))

            stop_response = await async_client.get(f"{stop_endpoint}/{last_id}")
            assert stop_response.status_code == 200
            assert stop_response.json() == {"status": "ok"}


async def _test_stop_inexistent_process(mocker, get_manager_func, start_endpoint, stop_endpoint):
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as async_client:
        async with override_dependency(mocker, get_manager_func) as manager:
            start_response = await start_process(async_client, start_endpoint, preset_end_status="loop")
            assert start_response.status_code == 201
            logger.debug("Processes: %s", manager.processes)

            inexistent_id = 9999

            stop_response = await async_client.get(f"{stop_endpoint}/{inexistent_id}")
            assert stop_response.status_code == status.HTTP_404_NOT_FOUND
            assert stop_response.json() == {
                "detail": "Process not found. It may have already exited or not started yet. Please check logs."
            }


# Test flows endpoints and interaction with db (read and write conf)
def test_flows(client):  # noqa: F811
    get_response = client.get("/api/v1/flows")
    assert get_response.status_code == 200
    data = get_response.json()["data"]
    assert "flows" in data

    response = client.post("/api/v1/flows", json=data)
    assert response.status_code == 200


# def test_get_build_status(client):
#     pass


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "end_status, process_status", [("failure", Status.FAILED), ("loop", Status.RUNNING), ("success", Status.COMPLETED)]
)
async def test_start_build(mocker, end_status, process_status):
    await _test_start_process(
        mocker,
        get_build_manager,
        endpoint="/api/v1/bot/build/start",
        preset_end_status=end_status,
        expected_end_status=process_status,
        timeout=BUILD_COMPLETION_TIMEOUT,
    )


@pytest.mark.asyncio
async def test_stop_build(mocker):
    await _test_stop_process(
        mocker, get_build_manager, start_endpoint="/api/v1/bot/build/start", stop_endpoint="/api/v1/bot/build/stop"
    )


@pytest.mark.asyncio
async def test_stop_build_bad_id(mocker):
    await _test_stop_inexistent_process(
        mocker, get_build_manager, start_endpoint="/api/v1/bot/build/start", stop_endpoint="/api/v1/bot/build/stop"
    )


# def test_get_run_status(client):
#     pass


# Test processes of various end_status + Test integration with get_status. No db interaction (mocked processes)
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "end_status, process_status", [("failure", Status.FAILED), ("loop", Status.RUNNING), ("success", Status.ALIVE)]
)
async def test_start_run(mocker, end_status, process_status):
    build_id = 43
    await _test_start_process(
        mocker,
        get_run_manager,
        endpoint=f"/api/v1/bot/run/start/{build_id}",
        preset_end_status=end_status,
        expected_end_status=process_status,
        timeout=RUN_RUNNING_TIMEOUT,
    )


@pytest.mark.asyncio
async def test_stop_run(mocker):
    build_id = 43
    await _test_stop_process(
        mocker,
        get_run_manager,
        start_endpoint=f"/api/v1/bot/run/start/{build_id}",
        stop_endpoint="/api/v1/bot/run/stop",
    )


@pytest.mark.asyncio
async def test_stop_run_bad_id(mocker):
    build_id = 43
    await _test_stop_inexistent_process(
        mocker,
        get_run_manager,
        start_endpoint=f"/api/v1/bot/run/start/{build_id}",
        stop_endpoint="/api/v1/bot/run/stop",
    )


@pytest.mark.asyncio
async def test_connect_to_ws(mocker):
    build_id = 43

    async with httpx.AsyncClient(transport=ASGIWebSocketTransport(app)) as client:
        async with override_dependency(mocker, get_run_manager) as process_manager:
            # Start a process
            start_response = await start_process(
                client,
                endpoint=f"http://localhost:8007/api/v1/bot/run/start/{build_id}",
                preset_end_status="success",
            )
            assert start_response.status_code == 201
            process_manager.check_status.assert_awaited_once()

            run_id = process_manager.get_last_id()
            logger.debug(f"run_id: {run_id}")
            await asyncio.sleep(10)

            assert await process_manager.get_status(run_id) == Status.ALIVE

            async with aconnect_ws(f"http://localhost:8000/api/v1/bot/run/connect?run_id={run_id}", client) as ws:
                message = await ws.receive_text()
                assert message == "Start chatting"


def test_search_service(client):
    get_response = client.get("/api/v1/services/search/is_upper_case")
    assert get_response.status_code == 200
    data = get_response.json()["data"]
    assert data


def test_get_conditions(client):
    get_response = client.get("/api/v1/services/get_conditions")
    assert get_response.status_code == 200
    data = get_response.json()["data"]
    assert data



# create test flows function here
import pytest
from omegaconf import OmegaConf

from chatsky_ui.api.api_v1.endpoints.flows import flows_get, flows_post


@pytest.mark.asyncio
async def test_flows_get(mocker):
    mocker.patch("chatsky_ui.api.api_v1.endpoints.flows.read_conf", return_value=OmegaConf.create({"foo": "bar"}))
    response = await flows_get()
    assert response["status"] == "ok"
    assert response["data"] == {"foo": "bar"}


@pytest.mark.asyncio
async def test_flows_post(mocker):
    mocker.patch("chatsky_ui.api.api_v1.endpoints.flows.write_conf", return_value={})
    response = await flows_post({"foo": "bar"})
    assert response["status"] == "ok"



"""
Process classes.
-----------------

Classes for build and run processes.
"""
import asyncio
import logging
import os
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

from dotenv import load_dotenv

from chatsky_ui.core.config import settings
from chatsky_ui.core.logger_config import get_logger, setup_logging
from chatsky_ui.db.base import read_conf, write_conf
from chatsky_ui.schemas.process_status import Status

load_dotenv()

GRACEFUL_TERMINATION_TIMEOUT = float(os.getenv("GRACEFUL_TERMINATION_TIMEOUT", 2))
PING_PONG_TIMEOUT = float(os.getenv("PING_PONG_TIMEOUT", 0.5))


def _map_to_str(params: Dict[str, Any]):
    for k, v in params.items():
        if isinstance(v, datetime):
            params[k] = v.strftime("%Y-%m-%dT%H:%M:%S")
        elif isinstance(v, Path):
            params[k] = str(v)


class Process(ABC):
    """Base for build and run processes."""

    def __init__(self, id_: int, preset_end_status: str = ""):
        self.id: int = id_
        self.preset_end_status: str = preset_end_status
        self.status: Status = Status.NULL
        self.timestamp: datetime = datetime.now()
        self.log_path: Path
        self.lock: asyncio.Lock = asyncio.Lock()
        self.process: asyncio.subprocess.Process  # pylint: disable=no-member #TODO: is naming ok?
        self.logger: logging.Logger

    async def start(self, cmd_to_run: str) -> None:
        """Starts an asyncronous process with the given command."""
        self.process = await asyncio.create_subprocess_exec(
            *cmd_to_run.split(),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.PIPE,
        )

    async def get_full_info(self, attributes: list) -> Dict[str, Any]:
        """
        Get the values of the attributes mentioned in the list.

        Args:
            attributes (list): A list of attributes to get the values of.

        Returns:
            dict: A dictionary containing the values of the attributes mentioned in the list.
        """
        await self.check_status()
        info = {key: getattr(self, key) for key in self.__dict__ if key in attributes}
        if "status" in attributes:
            info["status"] = self.status.value

        return info

    @abstractmethod
    async def update_db_info(self):
        raise NotImplementedError

    async def periodically_check_status(self) -> None:
        """Periodically checks the process status and updates the database."""
        while True:
            await self.update_db_info()  # check status and update db
            self.logger.info("Status of process '%s': %s", self.id, self.status)
            if self.status in [Status.STOPPED, Status.COMPLETED, Status.FAILED]:
                break
            await asyncio.sleep(2)  # TODO: ?sleep time shouldn't be constant

    async def check_status(self) -> Status:
        """Returns the process current status.

        Returns:
            - Status.NULL: When a process is initiated but not started yet. This condition is unusual and typically
                indicates incorrect usage or a process misuse in backend logic.
            - Status.ALIVE: process is alive and ready to communicate
            - Status.RUNNING: process is still trying to get alive. no communication
            - Status.COMPLETED: returncode is 0
            - Status.FAILED: returncode is 1
            - Status.STOPPED: returncode is -15
            - Status.FAILED_WITH_UNEXPECTED_CODE: failed with other returncode
        """
        if self.process is None:
            self.status = Status.NULL
        # if process is already alive, don't interrupt potential open channels by checking status periodically.
        elif self.process.returncode is None:
            if self.status == Status.ALIVE:
                self.status = Status.ALIVE
            else:
                if await self.is_alive():
                    self.status = Status.ALIVE
                else:
                    self.status = Status.RUNNING

        elif self.process.returncode == 0:
            self.status = Status.COMPLETED
        elif self.process.returncode == 1:
            self.status = Status.FAILED
        elif self.process.returncode == -15:
            self.status = Status.STOPPED
        else:
            self.logger.error(
                "Unexpected code was returned: '%s'. A non-zero return code indicates an error.",
                self.process.returncode,
            )
            self.status = Status.FAILED_WITH_UNEXPECTED_CODE

        if self.status not in [Status.NULL, Status.RUNNING, Status.ALIVE, Status.STOPPED]:
            stdout, stderr = await self.process.communicate()
            if stdout:
                self.logger.info(f"[stdout]\n{stdout.decode()}")
            if stderr:
                self.logger.error(f"[stderr]\n{stderr.decode()}")

        return self.status

    async def stop(self) -> None:
        """Stops the process.

        Raises:
            ProcessLookupError: If the process doesn't exist or already exited.
            RuntimeError: If the process has not started yet.
        """
        if self.process is None:  # Check if a process has been started
            self.logger.error("Cannot stop a process '%s' that has not started yet.", self.id)
            raise RuntimeError
        try:
            self.logger.debug("Terminating process '%s'", self.id)
            self.process.terminate()
            try:
                await asyncio.wait_for(self.process.wait(), timeout=GRACEFUL_TERMINATION_TIMEOUT)
                self.logger.debug("Process '%s' was gracefully terminated.", self.id)
            except asyncio.TimeoutError:
                self.process.kill()
                await self.process.wait()
                self.logger.debug("Process '%s' was forcefully killed.", self.id)
            self.logger.debug("Process returencode '%s' ", self.process.returncode)

        except ProcessLookupError as exc:
            self.logger.error("Process '%s' not found. It may have already exited.", self.id)
            raise ProcessLookupError from exc

    async def read_stdout(self) -> bytes:
        """Reads the stdout of the process for communication."""
        async with self.lock:
            if self.process is None:
                self.logger.error("Cannot read stdout from a process '%s' that has not started yet.", self.id)
                raise RuntimeError
            if self.process.stdout is None:
                raise RuntimeError(f"The process '{self.id}' stdout is None. It might be still running.")
            return await self.process.stdout.readline()

    async def write_stdin(self, message: bytes) -> None:
        """Writes a message to the stdin of the process for communication."""
        if self.process is None:
            self.logger.error("Cannot write into stdin of a process '%s' that has not started yet.", self.id)
            raise RuntimeError
        if self.process.stdin is None:
            raise RuntimeError(f"The process '{self.id}' stdin is None. It might be still running.")
        self.process.stdin.write(message)
        await self.process.stdin.drain()

    async def is_alive(self) -> bool:
        """Checks if the process is alive by writing to stdin andreading its stdout."""
        message = b"Hi\n"
        try:
            # Attempt to write and read from the process with a timeout.
            await self.write_stdin(message)
            output = await asyncio.wait_for(self.read_stdout(), timeout=PING_PONG_TIMEOUT)
            if not output:
                return False
            self.logger.debug("Process is alive and output afer communication is: %s", output.decode())
            return True
        except asyncio.exceptions.TimeoutError:
            self.logger.debug("Process is still running.")
            return False


class RunProcess(Process):
    """Process for running a DFF pipeline."""

    def __init__(self, id_: int, build_id: int, preset_end_status: str = ""):
        super().__init__(id_, preset_end_status)
        self.build_id: int = build_id

        self.log_path: Path = setup_logging("runs", self.id, self.timestamp)
        self.logger = get_logger(str(id_), self.log_path)

    async def get_full_info(self, attributes: Optional[list] = None) -> Dict[str, Any]:
        if attributes is None:
            attributes = ["id", "preset_end_status", "status", "timestamp", "log_path", "build_id"]
        return await super().get_full_info(attributes)

    async def update_db_info(self) -> None:
        # save current run info into runs_path
        self.logger.debug("Updating db run info")
        runs_conf = await read_conf(settings.runs_path)

        run_params = await self.get_full_info()
        _map_to_str(run_params)

        for run in runs_conf:
            if run.id == run_params["id"]:  # type: ignore
                for key, value in run_params.items():
                    setattr(run, key, value)
                break
        else:
            runs_conf.append(run_params)

        await write_conf(runs_conf, settings.runs_path)

        # save current run id into the correspoinding build in builds_path
        builds_conf = await read_conf(settings.builds_path)
        for build in builds_conf:
            if build.id == run_params["build_id"]:  # type: ignore
                if run_params["id"] not in build.run_ids:  # type: ignore
                    build.run_ids.append(run_params["id"])  # type: ignore
                    break

        await write_conf(builds_conf, settings.builds_path)


class BuildProcess(Process):
    """Process for converting a frontned graph to a DFF script."""

    def __init__(self, id_: int, preset_end_status: str = ""):
        super().__init__(id_, preset_end_status)
        self.run_ids: List[int] = []

        self.log_path: Path = setup_logging("builds", self.id, self.timestamp)
        self.logger = get_logger(str(id_), self.log_path)

    async def get_full_info(self, attributes: Optional[list] = None) -> Dict[str, Any]:
        if attributes is None:
            attributes = ["id", "preset_end_status", "status", "timestamp", "log_path", "run_ids"]
        return await super().get_full_info(attributes)

    async def update_db_info(self) -> None:
        # save current build info into builds_path
        builds_conf = await read_conf(settings.builds_path)

        build_params = await self.get_full_info()
        _map_to_str(build_params)

        for build in builds_conf:
            if build.id == build_params["id"]:  # type: ignore
                for key, value in build_params.items():
                    setattr(build, key, value)
                break
        else:
            builds_conf.append(build_params)

        await write_conf(builds_conf, settings.builds_path)
